use std::{collections::HashMap, error::Error};
use crate::reskit::soundtrack::{types::{PatternRow, Note}, formats::dmf::ChannelState};

// https://github.com/sikthehedgehog/Echo/blob/master/doc/esf.txt
const ESF_NOTE_ON: u8 = 0x00;
const ESF_NOTE_OFF: u8 = 0x10;
const ESF_SET_VOLUME: u8 = 0x20;
const ESF_SET_FREQUENCY: u8 = 0x30;
const ESF_SET_INSTRUMENT: u8 = 0x40;
const ESF_DELAY_SHORT: u8 = 0xD0;
const ESF_SFX_LOCK: u8 = 0xE0;
const ESF_SET_FM_PARAMS: u8 = 0xF0;
const ESF_SET_FM_DIRECT_0: u8 = 0xF8;
const ESF_SET_FM_DIRECT_1: u8 = 0xF9;
const ESF_SET_FLAGS: u8 = 0xFA;
const ESF_CLEAR_FLAGS: u8 = 0xFB;
const ESF_GOTO: u8 = 0xFC;
const ESF_LOOP_SET: u8 = 0xFD;
const ESF_DELAY_LONG: u8 = 0xFE;
const ESF_STOP: u8 = 0xFF;

pub const ESF_FM_1: u8 = 0x00;
pub const ESF_FM_2: u8 = 0x01;
pub const ESF_FM_3: u8 = 0x02;
pub const ESF_FM_4: u8 = 0x04;
pub const ESF_FM_5: u8 = 0x05;
pub const ESF_FM_6: u8 = 0x06;

pub const ESF_PSG_1: u8 = 0x08;
pub const ESF_PSG_2: u8 = 0x09;
pub const ESF_PSG_3: u8 = 0x0A;
pub const ESF_PSG_4: u8 = 0x0B;

pub const ESF_FM_6_PCM: u8 = 0x0C;

const ESF_SEMITONE_C: u8        = 0;
const ESF_SEMITONE_C_SHARP: u8  = 1;
const ESF_SEMITONE_D: u8        = 2;
const ESF_SEMITONE_D_SHARP: u8  = 3;
const ESF_SEMITONE_E: u8        = 4;
const ESF_SEMITONE_F: u8        = 5;
const ESF_SEMITONE_F_SHARP: u8  = 6;
const ESF_SEMITONE_G: u8        = 7;
const ESF_SEMITONE_G_SHARP: u8  = 8;
const ESF_SEMITONE_A: u8        = 9;
const ESF_SEMITONE_A_SHARP: u8  = 10;
const ESF_SEMITONE_B: u8        = 11;

pub type EchoEvent = Vec<u8>;

pub trait EchoFormat {

    fn get_eefs( &self ) -> Result<HashMap<String, Vec<u8>>, Box<dyn Error>>;

    fn get_eifs( &self ) -> Result<HashMap<String, Vec<u8>>, Box<dyn Error>>;

    fn get_ewfs( &self ) -> Result<HashMap<String, Vec<u8>>, Box<dyn Error>>;

    fn get_esf( &self ) -> Result<Vec<u8>, Box<dyn Error>>;
}

fn get_semitone( note: &Note ) -> Result<u8, Box<dyn Error>> {
    Ok( match note {
        Note::NoteOff => return Err( "internal error: attempted to get semitone for a Note::NoteOff" )?,
        Note::C(_) => ESF_SEMITONE_C,
        Note::CSharp(_) => ESF_SEMITONE_C_SHARP,
        Note::D(_) => ESF_SEMITONE_D,
        Note::DSharp(_) => ESF_SEMITONE_D_SHARP,
        Note::E(_) => ESF_SEMITONE_E,
        Note::F(_) => ESF_SEMITONE_F,
        Note::FSharp(_) => ESF_SEMITONE_F_SHARP,
        Note::G(_) => ESF_SEMITONE_G,
        Note::GSharp(_) => ESF_SEMITONE_G_SHARP,
        Note::A(_) => ESF_SEMITONE_A,
        Note::ASharp(_) => ESF_SEMITONE_A_SHARP,
        Note::B(_) => ESF_SEMITONE_B
    } )
}

fn get_fm_note_byte( octave: u8, semitone: u8 ) -> u8 {
    32 * octave + 2 * semitone + 1
}

fn get_psg_note_byte( octave: u8, semitone: u8 ) -> u8 {
    24 * octave + 2 * semitone
}

fn get_note( note: &Note, channel: u8 ) -> Result<u8, Box<dyn Error>> {
    Ok( match channel {
        ESF_FM_1..=ESF_FM_3 | ESF_FM_4..=ESF_FM_6 => get_fm_note_byte( note.get_octave()?, get_semitone( note )? ),
        ESF_PSG_1..=ESF_PSG_4 => get_psg_note_byte( note.get_octave()?, get_semitone( note )? ),
        _ => return Err( "internal error: invalid channel" )?
    } )
}

fn get_volume( channel: u8, volume: u8 ) -> Result<u8, Box<dyn Error>> {
    Ok( match channel {
        ESF_FM_1..=ESF_FM_3 | ESF_FM_4..=ESF_FM_6 => 0x7F - volume,
        ESF_PSG_1..=ESF_PSG_4 => 0x0F - volume,
        _ => return Err( "internal error: invalid channel" )?
    } )
}

pub fn get_events( row: &PatternRow, channel: u8, channel_state: &mut ChannelState ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = Vec::new();

    if let Some( note ) = &row.note {
        if let Note::NoteOff = note {
            events.push( vec![ ESF_NOTE_OFF | channel ] );
        } else {
            events.push( vec![ ESF_NOTE_ON | channel, get_note( note, channel )? ] );
        }
    }

    if let Some( volume ) = &row.volume {
        // https://github.com/sikthehedgehog/Echo/blob/master/doc/esf.txt#L206-L207
        events.push( vec![ ESF_SET_VOLUME | channel, get_volume( channel, *volume )? ] );
    }

    if let Some( instrument ) = &row.instrument_index {
        // Do not set the same instrument if it was already set before
        if &row.instrument_index != &channel_state.current_instrument {
            events.push( vec![ ESF_SET_INSTRUMENT | channel, *instrument as u8 ] );
            channel_state.current_instrument = Some( *instrument );
        }
    }

    Ok( events )
}